// Fichiers d'entête système.
#include <string>

#include "gl-utils.h"

// On place ces variables en global car on en aura besoin pour libérer la mémoire.
static GLuint program; // Le handle du programme des shaders.
static GLuint vbo;     // Le handle du VBO.
static GLuint vao;     // Le handle du VAO.

// Fonction d'initialisation, appelée une fois en début de programme.
static void init(const std::string &shader_name)
{
  program = read_shader(shader_name + ".vert", shader_name + ".frag");

  // Créations des buffers.
  GLint  attrib_location;
  GLfloat positions[] = { -0.5, -0.5,  0.5, -0.5,  0.0, 0.5 }; // 3 fois deux positions (x, y) pour dessiner un triangle.

  // Le VAO pour stoker les réglages.
  glGenVertexArrays(1, &vao);                 PRINT_OPENGL_ERROR();
  glBindVertexArray(vao);                     PRINT_OPENGL_ERROR();

  // Le VBO pour stoker les données des coordonnées.
  glGenBuffers(1, &vbo);                      PRINT_OPENGL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, vbo);         PRINT_OPENGL_ERROR();
  glBufferData(GL_ARRAY_BUFFER,
               sizeof positions, positions,
               GL_STATIC_DRAW);               PRINT_OPENGL_ERROR();

  // On indique le format des données.
  attrib_location = glGetAttribLocation(program, "position");          PRINT_OPENGL_ERROR();
  glVertexAttribPointer(attrib_location, 2, GL_FLOAT, GL_FALSE, 0, 0); PRINT_OPENGL_ERROR();
  glEnableVertexAttribArray(attrib_location);                          PRINT_OPENGL_ERROR();
  // La couleur d'effacement sera du gris-bleu.
  glClearColor(0.5, 0.6, 0.7, 1.0);                                    PRINT_OPENGL_ERROR();
}

// Fonction d'affichage, appelée régulièrement, pour réafficher le contenu.
static void display(void)
{
  GLint uniform_location;

  glClear(GL_COLOR_BUFFER_BIT);      PRINT_OPENGL_ERROR();

  uniform_location = glGetUniformLocation(program, "timer"); PRINT_OPENGL_ERROR ();
  if (uniform_location != -1)
    glUniform1f(uniform_location, glutGet(GLUT_ELAPSED_TIME));

  glDrawArrays(GL_TRIANGLES, 0, 3);  PRINT_OPENGL_ERROR();

  glutSwapBuffers();
}

// Fonction de libération des ressources, appélée en fin de programme.
static void terminate(void)
{
  glBindBuffer(GL_ARRAY_BUFFER, 0); PRINT_OPENGL_ERROR();
  glDeleteBuffers(1, &vbo);         PRINT_OPENGL_ERROR();
  glBindVertexArray(0);             PRINT_OPENGL_ERROR();
  glDeleteVertexArrays(1, &vao);    PRINT_OPENGL_ERROR();
  glUseProgram(0);                  PRINT_OPENGL_ERROR();
  glDeleteProgram(program);         PRINT_OPENGL_ERROR();
}

static void idle(void)
{
  glutPostRedisplay();
}

int main(int argc, char* argv[])
{
  std::string shader = "shaders/simple";
  glutInitContextVersion(3, 3);
  glutInitContextFlags(GLUT_FORWARD_COMPATIBLE | GLUT_DEBUG);
  glutInitContextProfile(GLUT_CORE_PROFILE);
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize(800, 600);
  glutCreateWindow("Un triangle tout bizarre");
  glewExperimental = true; // Pour supporter le profile core.
  glewInit(); // Initialisation de la bibliothèque GLEW.

  if (argc > 1) shader = argv[1];
  init(shader);

  glutIdleFunc(idle);
  glutDisplayFunc(display);

  glutMainLoop();

  terminate();

  return 0;
}
