#version 330

struct Object
{
  vec4 geometry; //!< center/radius for spheres, normal/distance for planes
  vec3 color;
  bool type;     //!< true for spheres, false for planes
};


struct Inter_data
{ 
  vec3 inter;
  vec3 normal;
  vec3 color;
  bool is_inter;
};

uniform float timer = 0.0;

const vec4 shad = vec4(0.2, 0.8, 0.8, 128.0);
uniform vec3 light = vec3(15.0, -15.0, -10);


//----------------------------------------
const vec3 cam = vec3(0, 0, -2.0);
const vec3 screen_0 = vec3(-1.0, -1.0, 0.0);
const vec3 screen_dx = vec3(2.0, 0.0, 0.0);
const vec3 screen_dy = vec3(0.0, 2.0, 0.0);
uniform float angle=0.0;
uniform vec3 screen_tr=vec3(0.0,0.0,0.0);

Object objects[4] = Object[4]
(
  Object(vec4(-0.5, 1.0, 2.0, 0.9), vec3(150.0/255.0, 150.0/255.0, 1.0), true),
  Object(vec4( 1.0, 0.8+sin(timer/2300.0), 1.0, 0.2), vec3(1.0,           0.0,       0.0), true),
  Object(vec4( 1.7,-0.2, 5.0, 0.3), vec3(100.0/255.0,   1.0,       0.0), true),
  Object(vec4( 0.0,-1.0, 0.0, 1.0), vec3(1.0, 1.0, 1.0), false)
);

in vec2 u;

out vec4 out_color;

bool plane_intersect(in vec4 plane, in vec3 x0, in vec3 dir,
                     out float t, out vec3 inter, out vec3 normal)
{
  float epsilon = 0.0001;
  float proj = dot(dir, plane.xyz);

  t = 0.0;
  inter = vec3(0.0);
  normal = plane.xyz;

  if (abs(proj) < epsilon)
    return false;

  t = -dot(plane, vec4(x0, 1.0)) / proj;
  inter = x0 + t * dir;
  return t > 0.0;
}

bool sphere_intersect(in vec4 sphere, in vec3 x0, in vec3 dir,
                      out float t, out vec3 inter, out vec3 normal)
{
  float B;
  float C;

  t = 0.0;
  inter = vec3(0.0);
  normal = vec3(0.0, 1.0, 0.0);
  B = 2.0 * dot(dir, x0 - sphere.xyz);
  C = dot(x0 - sphere.xyz, x0 - sphere.xyz) - sphere.w * sphere.w;
  if (B * B - 4.0 * C >= 0.0)
    {
      t = (-B - sqrt(B * B - 4.0 * C)) / 2.0;
      if (t > 0.0)
        {
          inter = x0 + t * dir;
          normal = normalize(inter - sphere.xyz);
          return true;
        }
      t = (-B + sqrt(B * B - 4.0 * C)) / 2.0;
      if (t > 0.0)
        {
          inter = x0 + t * dir;
          normal = normalize(sphere.xyz - inter);
          return true;
        }
    }
  return false;
}

bool throw_shadow_ray(in vec3 x0, in vec3 dir)
{
  float t;
  vec3 inter = vec3(0.0);
  vec3 normal = vec3(0.0, 1.0, 0.0);

  for (int i = 0; i < objects.length(); i++)
    {
      bool shadow = false;
      if (objects[i].type)
        shadow = sphere_intersect(objects[i].geometry, x0, dir, t, inter, normal);
      else
        shadow = plane_intersect(objects[i].geometry, x0, dir, t, inter, normal);
      if (shadow)
        return false;
    }
  return true;
}

Inter_data throw_ray(in vec3 x0, in vec3 dir)
{
  float min_t = 10000.0;
  vec3  color = vec3(0.0);

  Inter_data data=Inter_data(vec3(0.0,0.0,0.0),vec3(0.0,0.0,0.0),vec3(0.0,0.0,0.0),false);

  for (int i = 0; i < objects.length(); i++)
    {
      float t;
      vec3  inter = vec3(0.0);
      vec3  normal = vec3(0.0, 1.0, 0.0);
      bool  found = false;

      if (objects[i].type)
        found = sphere_intersect(objects[i].geometry, x0, dir, t, inter, normal);
      else
        found = plane_intersect(objects[i].geometry, x0, dir, t, inter, normal);
      if (found && t < min_t)
        {
          vec3 light_dir      = normalize(light - inter);
          vec3 view           = normalize(inter - x0);
          vec3 reflected_view = view - 2.0 * dot(view, normal) * normal;



          min_t = t;
          if (objects[i].type)
            color = objects[i].color;
          else
            color = vec3(float(mod(int(inter.x + timer/1700.0 + 0.5) + int(inter.y + 0.5) + int(inter.z + 0.5), 2)));
          if (throw_shadow_ray(inter + 0.0001 * normal, light_dir))
            {
              color *= shad.x + shad.y * max(0.0, dot(normal, light_dir)); // Add the diffuse part
              color += vec3(shad.z * pow(max(0.0, dot(reflected_view, light_dir)), shad.w)); // And the specular part.
            }
          else
            color *= shad.x;

          data=Inter_data(inter,normal,color,true);
        }
    }
  return data;
}

void main()
{
  vec3 screen_dx_current = u.x * screen_dx;
  vec3 screen_dy_current = u.y * screen_dy;

  mat3 rot=mat3(cos(angle),0,-sin(angle), 0,1,0, sin(angle),0,cos(angle));

  vec3 y = rot*(screen_0+ screen_dx_current + screen_dy_current-cam)+cam+screen_tr;
  vec3 ray_x0 = cam+screen_tr;
  vec3 ray_u = normalize(y -(cam+screen_tr));


  Inter_data d=throw_ray(ray_x0, ray_u);
  out_color=vec4(d.color,1.0);
  float dL=0.0001f;
  float attenuation=1.0f;
  for(int k=0;k<5;++k)
  {
    if(d.is_inter==true)
    {
      vec3 view = normalize(d.inter - ray_x0);
      ray_u=normalize(ray_u - 2.0 * dot(view, d.normal) * d.normal);
      ray_x0=d.inter;
      d=throw_ray(ray_x0+dL*ray_u,ray_u);
    
      attenuation*=0.2;
      out_color += attenuation*vec4(d.color,1.0);
    }
  }
}

